; LCD_4BIT_407.a51    LCD Display drive 4-BIT  
;  add ALE assertion (11/13)
;  Jan. 2007: TRY REVERSING ORDER of nybble-sends (4/07)
; test program puts out short message, on two lines
; 5/12/06 try correcting my erroneous idea that BUSY travels on a dedicated pin
;  it doesn't: it's d7 of the data port (P1.7, here)
; 5/14/06 try releasing BUSY bit (P1.7) before testing it with Read (drive it high)
; 5/23/06 modified to send first 3 init commands without checking busy, per Philips advice,
;  and send 3 times (weird)

; PORT ASSIGNMENTS:
; DATA: P1 (4 lines); P1.7 carries BUSY bit on a Read
; EN: P3.5, RS: P3.4, R/W* P3.3

    $NOSYMBOLS        ; keeps listing short

; default Ride7 path, Mar. 16
    $INCLUDE (C:\Program Files\Raisonance\Ride\inc\51\Dallas\REG320.INC) 
    $INCLUDE (C:\Program Files\Raisonance\Ride\inc\51\Dallas\vectors320.INC) 
	
	STACKBOT EQU 07Fh	; Not necessary, but a good place for it
	CR EQU 0Dh			; end-of-line code: CR
	EOT EQU 03h			; end-of-text code

; the following are PORT 1 DATA commands (defined in LCD instruction set)
	EIGHT_BIT EQU 030h	; these are LCD command codes (see data sheet p. 11ff) This is 3h, placed in 4 MSB's
	FOUR_BIT EQU 020h		; this is just a nybble: 2h, placed in 4 MSB's
	FUNC_SET EQU 28h		; 4-bit, dual line, 5X8 font (this is a byte)
	OFF EQU 08h
	CLEAR EQU 01h	 
	HOME EQU 03h		; try 3	(was 2) 6/1/06
	ENTRY EQU 06h     ; entry "mode": increment cursor after use;
							;  shift cursor, not display (changed to 06 from 07 (4/2007)
	ON EQU 0Eh			; display and cursor on
	CURSOR_OFF EQU 0Ch	; advances cursor one place
	NEXT_LINE EQU 0C0h	; to start of next line (I think--this from old HC05 listing, not data sheet); confirmed by 8052.com
				; command is 80h for set cursor to address, and remaining 4 bits define cursor address: 40h for start of 2d line
				; ...and this is PORT1 command, again, for LCD data lines
	PERMIT_BSY_RD EQU 0FFh	; output to P1, to allow input on all bits
	
	
; the following three are MY IDIOSYNCRATIC PORT 3 CONTROL bit patterns, driving EN, RS and R/W	
	COMMAND_WRITE EQU 0C7h	; bit pattern to Port 3, revised (5/18) to drive top 2 bits hi,
									;  since they = RD* and WR* on the bus
									;  and perhaps might cause trouble if asserted, though I don't think so
	COMMAND_READ EQU 0CFh		; ...to read BUSY flag
	

	DATA_WRITE EQU 0D7h		; ...ditto, for data write
	DELAYVAL EQU 028h		; delay value, for lazy initializations--before we can test BUSY bit
					;  this is more than required, but satisfies the 4ms required for longest wait
	
	
; bits:
	EN EQU P3.5
	RS EQU P3.4
	R_W EQU P3.3


;;;;;NOW PROGRAM CODE:
	ORG 0H
	LJMP BEGIN

        ORG 180h         

BEGIN:   MOV 0C4h, #04h	; turn on ALE during ROM run (bit 2, ALEON, of PMR register, C4h)
    ACALL LCDINIT		; go off and set things up: LCD
	MOV R0,#1					; init index to data table;

	
PUTOUT1:	ACALL GETCHAR		; go get ascii for one character
	CJNE A, #CR, DOIT		; if not yet end of line, send character 
	SJMP LINE2				; if end of line, go to next line
DOIT: ACALL WDATA			; go put out the character
	INC R0					; point to next character
	SJMP PUTOUT1			; ...keep putting out those characters
	
LINE2: MOV A, #NEXT_LINE	; command to go to next line, I think
	ACALL WCTRL
	INC R0					; advance index, so as not to print CR
	
PUTOUT2:		ACALL GETCHAR			; go get next character
	CJNE A, #EOT,DOMORE	; check for end of whole message
	MOV A, #CURSOR_OFF	; turn off unsightly cursor
	ACALL WCTRL
HALT:		SJMP HALT		; end of whole program: stop here

DOMORE: ACALL WDATA
	INC R0
	SJMP PUTOUT2			; keep pumping them out till EOT sensed
	

;---SUBROUTINES----

LCDINIT:  CLR EN
	MOV R7,#4					; do this command 3 times (very strange!) (1/19/07 changed from 4 to 3; back, 4/07))
	
	MOV P3, #COMMAND_WRITE	; set up controls for commands to initialize LCD
	MOV P1, #EIGHT_BIT	; this is an oddball: must send without first checking BUSY 
FOUR_X:	ACALL PULSE
	ACALL DELAY					; wait about 5 ms (4.1ms min. is spec'd for first of these writes)
	DJNZ R7,FOUR_X		; after this loop, display should be initialized to deliver its BUSY flag on d7
	
	MOV P1, #FOUR_BIT		; this is the last 4-bit output (a 2) before we start sending bytes
	ACALL PULSE
	ACALL DELAY
	
;   MOV A, #FOUR_BIT		; get byte (though I think this is wrong--just a nybble expected)
;  ACALL WCTRL	

	MOV A, #FUNC_SET			; ...so, we now check BUSY before sending anything more
	ACALL WCTRL	

	MOV A, #OFF			
	ACALL WCTRL

	MOV A, #CLEAR
	ACALL WCTRL
	
	MOV A, #ENTRY		; right-shift
	ACALL WCTRL	
	
;	MOV A, #ON
;	ACALL WCTRL		
	
	RET

;---this tiny routine just sends, as COMMAND, whaever is pased it in Acc
; this is FOUR-BIT VERSION: differs from 8-bit in sending in 2 passes, using SWAP A
WCTRL: MOV P3, #COMMAND_READ		; set up control signals
	MOV P1, #PERMIT_BSY_RD		; release all data lines, to allow Read (drive them high)
	SETB EN				; send EN high to allow reading BUSY flag
	
TEST1:	JB ,P1.7,TEST1			; wait till not busy
	CLR EN						; release EN, terminating Read operation
;	SETB P1.7					; drive msb hi, to permit later write (NONSENSE?)
	MOV P3, #COMMAND_WRITE
	swap a						; TRY SENDING LOW 4 FIRST (4/07)	
	MOV P1, A					; Now send the command (only the high 4 bits will go)
	ACALL PULSE					; ...then pulse enable to clock it in
	SWAP A						; ...interchange top and bottom nybbles...
	MOV P1, A					; Now send the command (now the low 4 bits will go)
	ACALL PULSE					; ...then pulse enable to clock it in
	SETB R_W						; disassert Write*
	RET
	
;---this sends, as DATA, whaever is pased it in Acc
; this is FOUR-BIT VERSION: differs from 8-bit in sending in 2 passes, using SWAP A
WDATA: MOV P3, #COMMAND_READ		; set up to check Busy flag (ndd in simulation)
	MOV P1, #PERMIT_BSY_RD	; release all data lines, to allow Read (drive them high)
	SETB EN						; send EN high to allow reading BUSY flag
TEST2:	JB ,P1.7,TEST2	
	CLR EN						; release EN, terminating Read operation
;	SETB P1.7					; drive msb hi, to permit later write (NONSENSE?)

	MOV P3, #DATA_WRITE		; now set up data write
;	swap a						; TRY LOW 4 FIRST (4/07)
	MOV P1, A					; ...and send the data (top 4 bits, first)...
	ACALL PULSE					; ...then pulse enable to clock it in
	SWAP A
	MOV P1, A					; Now send the data (now the low 4 bits)
	ACALL PULSE					; ...then pulse enable to clock it in	
	SETB R_W					; disassert Write*
	RET	
	
;--brief pulse to EN--
PULSE: SETB EN
	NOP					; make pulse last at least 450 nseconds
	CLR EN
	RET


GETCHAR: MOV A, R0		; recall saved index value, for use in indexed read
;			INC A				; move index beyond the RET (1/19/07)
			MOVC A, @A+PC	; 	pick up table entry
			RET

			DB 'Seems to work!'
			DB CR
			DB 'Is this second?'
			DB EOT
			
	
DELAY:	PUSH ACC				; save registers that'll get messed up
			PUSH B
			MOV A,#DELAYVAL	; get outer-loop delay value (# of 256-DJNZ loops)
INITINNER:	MOV B,#0			; initialize inner loop counter
INLOOP:	DJNZ B, INLOOP		; count down inner loop, till inner hits zero
			DJNZ ACC,INITINNER	; ...then dec outer, and start inner again.
			POP B  				; restore saved registers
			POP ACC
			RET					; Now back to main program.
		

	END
